"""
Test hypy.lib module
"""

from __future__ import with_statement

import unittest
import os
from contextlib import contextmanager

import hypy

from hypy import (HDocument, HDatabase, HHit, HResults, HCondition, OpenFailed,
        PutFailed, CloseFailed, FlushFailed, EditFailed)

class TestHDocument(unittest.TestCase):
    """
    Test the dictionary and text properties of HDocument
    """
    def setUp(self):
        """
        Create a document
        """
        self.doc = HDocument(uri=u'1')

    def test_dictlike(self):
        """
        HDocument mostly conforms to the dictionary protocol.  Make sure that
        works. 
        """
        doc = self.doc

        # byte strings, other types are not allowed.
        self.assertRaises(TypeError, doc.__setitem__, 'foobar', 'baz')
        self.assertRaises(TypeError, doc.__setitem__, 'foobar', 1)

        doc[u'foobar'] = u'baz'
        doc[u'foobar']
        self.assertEqual(doc[u'foobar'], u'baz')
        self.assertEqual(doc.get(u'foobar', 'default'), u'baz')
        self.assertEqual(doc.get(u'xyz', 'default'), 'default')
        self.assertEqual(doc.get(u'xyz'), None)

        newattrs = {u'new1': u'lala', u'foobar': u'bazz'}
        doc.update(newattrs)
        self.assertEqual(sorted(doc.items()), [(u'@uri', u'1'), (u'foobar', u'bazz'), (u'new1',
            u'lala')])

        doc[u'ninjas'] = u'11'
        self.assertEqual(sorted(doc.keys()), [u'@uri', u'foobar', u'new1', u'ninjas', ])
        self.assertEqual(sorted(doc.values()), [u'1', u'11', u'bazz', u'lala'])

        self.assertRaises(NotImplementedError, doc.__delitem__, u'ninjas')

    def test_text(self):
        """
        Mess with document text
        """
        doc = self.doc
        self.assertRaises(TypeError, doc.addText, 'xyz')
        doc.addText(u'xyz')
        self.assertEqual([u'xyz'], doc.getTexts())
        doc.addText(u'123')
        self.assertEqual([u'xyz', u'123'], doc.getTexts())
        self.assertRaises(TypeError, doc.addHiddenText, 'abc')
        doc.addHiddenText(u'abc')
        self.assertEqual([u'xyz', u'123'], doc.getTexts())

        self.assertEqual(u'xyz\n123', doc.text)

        doc.addText(u'\u062b')

        self.assertEqual('xyz123\xd8\xab', doc.encode('utf-8'))

    def test_unicodeType(self):
        """
        Almost everything in hypy must be unicode
        """
        self.assertRaises(TypeError, HDocument, uri='notunicode')


class TestDatabase(unittest.TestCase):
    """
    Tests HResults, HCondition and HHit.  And since you can't test these
    things without a database, test HDatabase.
    """
    @contextmanager
    def freshenDatabase(self, extras=0):
        """
        Use:
            with self.freshenDatabase() as db:
                ... stuff that should test using these three documents ...
        """
        db = HDatabase()
        db.open('_temp_db', 'w')

        doc = HDocument(uri=u'1')
        doc.addText(u'word this is my document. do you like documents? this one is hi-res.')
        db.putDoc(doc, clean=True)

        doc = HDocument(uri=u'2')
        doc.addText(u'word lorem ipsum dolor sit amet something whatever whatever i do not remember the rest')
        db.putDoc(doc)

        doc = HDocument(uri=u'3')
        doc.addText(u'word four score and 7 years ago our forefathers brought forth upon upon something')
        db.putDoc(doc)

        for x in range(4, extras+4):
            doc = HDocument(uri=unicode(x))
            doc.addText(u'filler filler filler carrot top')
            # set some attributes for attribute operator testing
            doc[u'specialId'] = unicode(x)
            doc[u'description'] = unicode(x) * x
            doc[u'date'] = u'2008-12-%s' % (x,)
            db.putDoc(doc)

        db.flush()

        try:
            yield db
        finally:
            try:
                db.close()
            except CloseFailed:
                """Some of the tests do this close on their own."""

    def test_dbOptimize(self):
        """
        Make sure the various optimize flags do not cause a heart attack.
        """
        with self.freshenDatabase() as db:
            db.optimize(purge=True)
        with self.freshenDatabase() as db:
            db.optimize(opt=True)
        with self.freshenDatabase() as db:
            db.optimize()
            self.assertRaises(NotImplementedError, db.sync)

    def test_removeUpdate(self):
        """
        Test for document id, update, document removal, len() of database.
        """
        docxx = HDocument(uri=u'xx')
        docxx.addText(u'xx')

        # id of a non-stored document
        self.assertEqual(docxx.id, -1)

        # updateAttributes on opened db?
        db = HDatabase()
        self.assertRaises(EditFailed, db.updateAttributes, docxx)

        # removes and updates
        with self.freshenDatabase() as db:
            # flags; just test that these do not nuke us. no idea what they
            # are supposed to do.
            db.putDoc(docxx, clean=True)
            del db[u'xx']
            # delete same doc twice?
            self.assertRaises(EditFailed, db.remove, uri=u'xx')

            db.putDoc(docxx)

            # __len__
            self.assertEqual(len(db), 4)

            # remove by uri, by id, by reference
            db.remove(uri=u'1')
            self.assertEqual(len(db), 3)
            doc2 = db[u'2']
            db.remove(doc2)
            self.assertEqual(len(db), 2)
            doc3id = db[u'3'].id
            db.remove(id=doc3id)
            self.assertEqual(len(db), 1)
            # no arg?
            self.assertRaises(TypeError, db.remove)
            # already removed?
            self.assertRaises(EditFailed, db.remove, id=doc3id)

            self.assertRaises(KeyError, db.__getitem__, '1')

            # fetch a document from the database, edit it, store it, compare
            # it with the original (unfetched) document.  Verify that they are
            # different after the edit. Then verify that the document can be
            # fetched (again) from the database with the edited text.
            dbdocxx = db[u'xx']
            # yes, these are different objects
            self.assertFalse(docxx is dbdocxx)
            self.assertTrue(dbdocxx.get(u'zz') is None)

            dbdocxx[u'zz'] = u'hello'
            db.updateAttributes(dbdocxx)
            dbdocxx2 = db[u'xx']
            # again, different objects
            self.assertFalse(dbdocxx is dbdocxx2)
            self.assertEqual(dbdocxx2.get(u'zz'), u'hello')

    def test_removeURINone(self):
        """
        #356253: should be able to explicitly say "uri=None" when calling remove
        """
        with self.freshenDatabase() as db:
            # this is odd, but it should work
            db.remove(id=1, uri=None)

    def test_removeNulls(self):
        """
        Bug 321579: nulls should not kill addText
        """
        docnulls = HDocument(uri=u'nulls\0')
        docnulls.addText(u'hello there\0 children')
        docnulls.addHiddenText(u'sweet\0sweet love')
        docnulls[u'character\0'] = u'chef\0'
        with self.freshenDatabase() as db:
            db.putDoc(docnulls)
            db.flush()
            cond = HCondition(u'there*')
            self.assertEqual(db.search(cond).pluck(u'@uri'), [u'nulls'])
            cond = HCondition(u'sweet*')
            self.assertEqual(db.search(cond).pluck(u'@uri'), [u'nulls'])

            cond = HCondition()
            cond.addAttr(u'character STREQ chef')
            self.assertEqual(db.search(cond).pluck(u'@uri'), [u'nulls'])

            docnulls[u'voi\0ce'] = u'Isaac Hayes'
            db.updateAttributes(docnulls)

            cond = HCondition()
            cond.addAttr(u'voice STRRX .saac.*')
            self.assertEqual(db.search(cond).pluck(u'@uri'), [u'nulls'])

    def test_putFlags(self):
        """
        Tests for put flags, other put-related corner cases.
        """
        docxx = HDocument(uri=u'xx')
        docxx.addText(u'xx')

        with self.freshenDatabase() as db:
            # flags; just test that these do not nuke us. no idea what they
            # are supposed to do.
            db.putDoc(docxx, clean=True)
            del db[u'xx']
            db.putDoc(docxx, weight=True)
            del db[u'xx']

            ## # put same doc twice?
            ## apparently this works. huh.
            ## db.putDoc(docxx); db.putDoc(docxx)

    def test_condExtras(self):
        """
        Tests for search skip, search options, cond on attributes
        """
        with self.freshenDatabase(extras=10) as db:
            self.assertEqual(len(db), 13)
            # skip and max
            cond4_8 = HCondition(u'filler', max=5)
            cond9_11 = HCondition(u'filler', max=3, skip=5)
            res1 = db.search(cond4_8)
            self.assertEqual(res1.pluck(u'@uri'), list(u'45678'))
            res2 = db.search(cond9_11)
            self.assertEqual(res2.pluck(u'@uri'), [u'9', u'10', u'11'])

            # union matching
            result = db.search(HCondition(u'ipsum score', matching='simple'))
            self.assertEqual(len(result), 0)
            result = db.search(HCondition(u'ipsum score', matching='union'))
            self.assertEqual(len(result), 2)

            # isect matching
            result = db.search(HCondition(u'lorem* ipsum*', matching='simple'))
            self.assertEqual(len(result), 1)
            result = db.search(HCondition(u'lorem* ipsum*', matching='isect'))
            self.assertEqual(len(result), 0)
            result = db.search(HCondition(u'lorem ipsum', matching='isect'))
            self.assertEqual(len(result), 1)

            # rough matching
            result = db.search(HCondition(u'lorem* ipsum*', matching='simple'))
            self.assertEqual(len(result), 1)
            result = db.search(HCondition(u'lorem* ipsum*', matching='rough'))
            self.assertEqual(len(result), 0)
            result = db.search(HCondition(u'lorem ipsum', matching='rough'))
            self.assertEqual(len(result), 1)

            # fewer-than-max hits
            result = db.search(HCondition(u'7*', matching='simple', max=2))
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0][u'@uri'], u'3')

            # attribute conditions
            cc = HCondition(u'*.**')
            self.assertRaises(TypeError, cc.addAttr, 'xxx')

            # None as a kwarg should work (bug #356253)
            c2 = HCondition(phrase=None)

            def attrSearch(expr, order=None, phrase=None):
                if phrase:
                    cond = HCondition(phrase)
                else:
                    cond = HCondition()
                cond.addAttr(expr)
                if order:
                    self.assertRaises(TypeError, cond.setOrder, '*.**')
                    cond.setOrder(order)
                return db.search(cond)
            result = attrSearch(u'description STREQ 4444')
            self.assertEqual(result.pluck(u'@uri'), [u'4'])
            result = attrSearch(u'specialId NUMGE 10')
            self.assertEqual(result.pluck(u'@uri'), [u'10', u'11', u'12', u'13'])
            result = attrSearch(u'date NUMGE 2008-12-09')
            self.assertEqual(sorted(map(int, result.pluck(u'@uri'))), [9, 10, 11, 12, 13])
            result = attrSearch(u'description STRRX .{10,14}')
            self.assertEqual(result.pluck(u'@uri'), [u'10', u'11', u'12', u'13'])
            # ordering and !not
            result = attrSearch(u'description !STRRX .{10,14}', u'@uri NUMA')
            self.assertEqual(result.pluck(u'@uri'), list(u'123456789'))
            result = attrSearch(u'description STRRX .{10,14}', u'@uri NUMD')
            self.assertEqual(result.pluck(u'@uri'), [u'13', u'12', u'11', u'10'])

            # one attribute search combined with a text search - regular
            # expression match, all characters
            result = attrSearch(u'description STRRX .{10,14}', u'@uri NUMD', u'*.**')
            self.assertEqual(result.pluck(u'@uri'), [u'13', u'12', u'11', u'10'])

    def test_dbOpenClosed(self):
        """
        Tests for all the db open/close modes
        """
        docyy = HDocument(uri=u'yy')
        docyy.addText(u'yy')
        docxx = HDocument(uri=u'xx')
        docxx.addText(u'xx')
        condxx = HCondition(u'xx')
        condyy = HCondition(u'yy')

        db = HDatabase()
        # open of unreachable directory
        self.assertRaises(OpenFailed, db.open, 'does/not/exist', 'a')
        # close before successful open
        self.assertRaises(CloseFailed, db.close)

        # w mode
        db.open('_temp_db', 'w')
        self.assert_(os.path.exists('_temp_db/_idx'))
        db.putDoc(docyy)
        db.close()

        # r mode
        db.open('_temp_db', 'r')
        # write to read-only db
        self.assertRaises(PutFailed, db.putDoc, docxx)
        self.assertEqual(len(db.search(condyy)), 1)
        db.close()

        # a mode
        db.open('_temp_db', 'a')
        db.putDoc(docxx)
        db.flush()
        self.assertEqual(len(db.search(condxx)), 1)
        self.assertEqual(len(db.search(condyy)), 1)
        db.close()

        # w mode again - check that the db is clobbered
        db.open('_temp_db', 'w')
        self.assertEqual(len(db.search(condxx)), 0)
        db.close()

        # close after successful close
        self.assertRaises(CloseFailed, db.close)
        self.assertRaises(FlushFailed, db.flush)

    def test_queries(self):
        """
        Test various conditions against an index to make sure search works.
        """
        with self.freshenDatabase() as db:
            # plain search, 8-bit str
            result = db.search(HCondition(u'wor*', matching='simple'))
            self.assertEqual(len(result), 3)

            # unicode searches
            result = db.search(HCondition(u'wor*', matching='simple'))
            self.assertEqual(len(result), 3)

            # test simple query with multiple hits
            result = db.search(HCondition(u'res*', matching='simple'))
            self.assertEqual(result.pluck(u'@uri'), [u'1', u'2'])

            # vary query terms to check result scoring
            result = db.search(HCondition(u'someth* | whatever*', matching='simple', max=2))
            self.assertEqual(result.pluck(u'@uri'), [u'2', u'3'])
            result = db.search(HCondition(u'someth* | upon*', matching='simple', max=2))
            self.assertEqual(result.pluck(u'@uri'), [u'3', u'2']) # FIXME

            self.assertEqual(result.hintWords(), [u'someth', u'upon'])

    def test_hits(self):
        """
        Poke at the hits returned by a search and see if document data and
        teaser text come out right.
        """
        with self.freshenDatabase() as db:
            cc = HCondition(u'hi-res')
            for hit in db.search(cc):
                self.assertEqual(str(hit), 
'@digest=17de33c57e358f0fc5d57cd26a08b48e\n@id=1\n@uri=1\n\nword this is my document. do you like documents? this one is hi-res.\n')
                self.assertEqual(hit.teaser([u'document']), u'word this is my <strong>document</strong>. do you li ... ke <strong>document</strong>s? this one is hi-re ... ')
                self.assertEqual(hit.teaser([u'document'], 'rst'), u'word this is my **document**. do you li ... ke **document**s? this one is hi-re ... ')
                self.assertRaises(TypeError, hit.teaser, ['document'])
                self.assertRaises(NotImplementedError, hit.teaser, 
                        [u'document'], 'pdf')

    def test_autoflush(self):
        """
        Verify that autoflush is not on when it's not turned on, and that
        words are autoindexed when it is turned on
        """
        cond = HCondition(u'doc*')

        @contextmanager
        def setup(**kw):
            db1 = HDatabase(**kw)
            db1.open('_temp_db', 'w')

            doc = HDocument(uri=u'1')
            doc.addText(u'word this is my document. do you like documents? this one is hi-res.')
            db1.putDoc(doc, clean=True)

            try:
                yield db1
            finally:
                db1.close()

        # non-autoflush: can't search for doc*
        with setup() as db:
            self.assertEqual(len(db.search(cond)), 0)

        # autoflush: now you can.
        with setup(autoflush=True) as db:
            self.assertEqual(len(db.search(cond)), 1)


class TestMisc(unittest.TestCase):
    """
    Test misc. features of the package such as version string
    """
    def test_version(self):
        """
        __version__ and other release info can be found in copyright.py and
        __init__.py

        This doesn't really verify that copyright.py is generated correctly,
        just that it exists and has the right contents.
        """
        from hypy import __version__ as version1
        from hypy.copyright import __version__ as version2
        self.assertEqual(version1, version2)


if __name__ == '__main__':
    unittest.main()



"""
Hypy is copyright (c) 2009 Cory Dodt.

GENERATED FILE -- DO NOT EDIT
"""
from datetime import date

author = "Cory Dodt"
releaseDate = "2014-07-27"
releaseRevision = "b480ff576daa tip"
__version__ = "0.8.5.1"



import sys
sys.path.insert(0, '/home/cdodt/wc/Hypy')
from hypy import *

INDEX = 'breakfast/'
db = HDatabase()
db.open(INDEX, 'w') # create it, destroying old one if it exists
db.close()
db.open(INDEX, 'a')

doc = HDocument(uri=u'http://estraier.gov/example.txt')
doc.addText(u"Hello there, this is my text.")
db.putDoc(doc)

# db = HDatabase(autoflush=True)  ## or ...
db.autoflush = True
# turning on autoflush does not immediately flush, so do that.
db.flush()

doc2 = HDocument(uri=u'http://estraier.gov/pricelist.txt')
doc2.addText(u"""Coffee: $2.00
Toast: $1.00
Eggs (2, any style): $3.00          Eggs (9, any style): $13.50
Eggs (3, any style): $4.50          Eggs (10, any style): $15.00
Eggs (4, any style): $6.00          Eggs (11, any style): $16.50
Eggs (5, any style): $7.50          Eggs (12, any style): $18.00
Eggs (6, any style): $9.00          Eggs (13, any style): $19.50
Eggs (7, any style): $10.50         Eggs (14, any style): $21.00
Eggs (8, any style): $12.00         Eggs (15, any style): $22.50
...
with apologies to the New Yorker for this joke""")
doc2[u'maxprice'] = u'22.50'
doc2[u'minprice'] = u'1.00'
db.putDoc(doc2)

if 0:
    print doc[u'@id']
    ## prints 1
    print doc[u'@uri']
    ## prints http://estraier.gov/example.txt

db.remove(id=1)  
db.putDoc(doc) # put it back so we can remove it again :)
db.remove(doc)
db.putDoc(doc) # and again..
db.remove(uri=u'http://estraier.gov/example.txt')

if 0:
    # n.b. the document gets a NEW ID each time we put it in..
    print doc[u'@id']
    ## prints 4

    print len(db)
    ## prints 1

    print db[u'http://estraier.gov/pricelist.txt']

    print doc2.encode('utf-16')


doc5 = HDocument(u'http://estraier.gov/weighted.txt')
doc5.addText(u"This is my boom-stick.")
doc5.addHiddenText(u"eggs " * 30)
db.putDoc(doc5)

if 0:
    print doc5.encode('utf-8')
    # prints This is my boom-stick.

cond = HCondition(u'eggs')
results = db.search(cond)
if 0:
    for doc in results:
        print doc[u'@id']
    # prints 2, then 5

    # you can also do this another way, using pluck:

    print results.pluck(u'@id')
    # prints [u'5', u'2']

cond = HCondition(u'egg*')
results = db.search(cond)
if 0:
    print len(results)
    print results[0][u'@uri']

doc6 = HDocument(u'http://estraier.gov/spam.txt')
doc6.addText(u'spam and eggs')
db.putDoc(doc6)

if 0:
    # simple, intersection:
    print db.search(HCondition(u'spam* eggs*')).pluck(u'@id')
    # prints [u'6']

    # union
    print db.search(HCondition(u'eggs spam', matching='union')).pluck(u'@id')
    # prints [u'5', u'2', u'6']

    # unions with simple matching - you cannot use wildcard matches with
    # matching='union' but you can do so with '|' syntax
    print db.search(HCondition(u'egg* | spam')).pluck(u'@id')
    # prints [u'5', u'2', u'6']

words = [u'toast']
results = db.search(HCondition(u' '.join(words)))
hit = results[0]
if 0:
    print hit.teaser(words) # default is 'html'
    # prints ......

    # another way to get the original search words:
    words = results.hintWords()

    print hit.teaser(words, format='rst')
    # prints ......


cond = HCondition()
cond.addAttr(u'@id STREQ 5')

if 0:
    print db.search(cond)[0][u'@id']
    # prints 5

cond = HCondition()
cond.addAttr(u'@id STREQ 5')
if 0:
    print db.search(cond)[0][u'@id']
    # prints 5

# for clarity in the rest of these examples, I'll use this function:
def attrSx(expr):
    cond = HCondition()
    cond.addAttr(expr)
    return u' '.join(db.search(cond).pluck(u'@id'))

# give another document some interesting attributes
doc6[u'maxprice'] = u'100'
doc6[u'minprice'] = u'0'
doc5[u'date'] = u'2009-01-01'
doc6[u'date'] = u'2009-02-02'

# commit the documents again, so they can be searched.  Note that this has
# nothing to do with flush() - these changes only exist in memory until
# you commit them with putDoc() and flush() too.  In this case, we are
# still using autoflush.
db.putDoc(doc6)
db.putDoc(doc5)

if 0:
    # numeric searches

    print attrSx(u'maxprice NUMGE 50')
    # prints 6

    # NOTE: documents that don't have an attribute are treated as having it
    # set to "0" when doing a numeric comparison. therefore all the documents
    # match this one:
    print attrSx(u'minprice NUMLE 50')
    # prints 2 6 5
    # and two documents match this one:
    print attrSx(u'minprice NUMLE 0.99')
    # prints 2 5


    # date comparisons are numeric comparisons
    print attrSx(u'date NUMGE 2008-12-31')
    # prints 6 5
    print attrSx(u'date NUMGE 2009-01-30')
    # prints 6

    # regular expressions, why not

    print attrSx(u'@uri STRRX (pricelist.txt|spam.txt)')
    # prints 2 6

    # you can invert them, too, to get only documents not matching
    print attrSx(u'@uri !STRRX (pricelist.txt|spam.txt)')
    # prints 5

# and of course, you can use attribute searches in the same condition with
# phrase searches
cond = HCondition(u'spam')
cond.addAttr(u'minprice NUMLE 50')
if 0:
    print db.search(cond)[0][u'@id']
    # prints 6

if 0:
    print db.search(HCondition(u'e*')).pluck(u'@id')
    # prints [u'5', u'2', u'5']
    print db.search(HCondition(u'e*', max=2)).pluck(u'@id')
    # prints [u'5', u'2'] .. what did you expect?
    print db.search(HCondition(u'e*', skip=2)).pluck(u'@id')
    # prints [u'6']

cond = HCondition(u'e*')

if 0:
    # natural order
    print db.search(cond).pluck(u'@id')
    # prints [u'5', u'2', u'6']

    # numeric ascending
    cond.setOrder(u'@id NUMA')
    print db.search(cond).pluck(u'@id')
    # prints [u'2', u'5', u'6']

    # numeric descending
    cond.setOrder(u'@id NUMD')
    print db.search(cond).pluck(u'@id')
    # prints [u'6', u'5', u'2']




"""
Put a Pythonic face on estraiernative
"""
import sys
from HTMLParser import HTMLParser
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from _estraiernative import (Condition as CCondition,
    Database as CDatabase, Document as CDocument, EstError as CError)

class FlushFailed(Exception):
    """
    Could not remove the specified doc from the database.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class EditFailed(Exception):
    """
    Could not edit the specified doc in the database
    """
    def __init__(self, id, msg):
        self.id = id
        self.msg = msg

    def __str__(self):
        return 'Document %s: %s' % (self.id, self.msg)

class PutFailed(Exception):
    """
    Could not add the specified doc to the database.
    """
    def __init__(self, uri, msg):
        self.uri = uri
        self.msg = msg

    def __str__(self):
        return 'Document %s: %s' % (self.uri, self.msg)

class OpenFailed(Exception):
    """
    Could not open the database with the specified mode.
    """
    def __init__(self, *args):
        self.msg = args[0]

    def __str__(self):
        return self.msg

class CloseFailed(Exception):
    """
    Could not close the database.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


def enforceUnicode(s, argName, exceptionClass, encoding='utf8'):
    """
    Force s to be a unicode object or None; return a byte string
    """
    if s is None:
        return None
    if not type(s) is unicode:
        raise exceptionClass('%s must be unicode' % (argName,))
    return s.encode(encoding)

def unicodeToByte(argSpecs, exceptionClass):
    """
    Test each of the argNames to see whether they are unicode text objects.
    If they are not, raise exceptionClass
    """
    def decorator(fn):
        def wrapper(*a, **kw):
            a = list(a)
            for pos, name in argSpecs:
                if name in kw:
                    kw[name] = enforceUnicode(kw[name], name, exceptionClass)
                else:
                    if len(a) >= pos + 1:
                        a[pos] = enforceUnicode(a[pos], name, exceptionClass)
            return fn(*a, **kw)
        return wrapper

    return decorator


class HCondition(object):
    """
    A search condition.
    Use matching='simple', 'rough', 'union' or 'isect'
    """
    @unicodeToByte([(1,'phrase')], TypeError)
    def __init__(self, phrase=None, matching='simple', max=None, skip=None):
        self.condition = CCondition()
        if phrase is not None:
            self.condition.set_phrase(phrase)
        if max is not None:
            self.condition.set_max(max)
        if skip is not None:
            self.condition.set_skip(skip)
        flags = 0
        if matching == 'simple':
            flags |= CCondition.SIMPLE
        elif matching == 'rough':
            flags |= CCondition.ROUGH
        elif matching == 'union':
            flags |= CCondition.UNION
        elif matching == 'isect':
            flags |= CCondition.ISECT
        self.condition.set_options(flags)

    @unicodeToByte([(1,'expression')], TypeError)
    def addAttr(self, expression):
        """
        Use 'expression' to filter results by attributes
        """
        self.condition.add_attr(expression)

    @unicodeToByte([(1,'expression')], TypeError)
    def setOrder(self, expression):
        """
        Use 'expression' to order results by attributes
        """
        self.condition.set_order(expression)


class HDatabase(object):
    """
    A more pythonic interface to estraier's database

    With autoflush=True (the default), automatically flush after every
    document add, which simplifies indexing.

    Set autoflush=False to manually flush, which allows better performance
    when indexing lots of documents at once.

    Per http://hyperestraier.sourceforge.net/uguide-en.html Hyper Estraier
    expects data to be stored as utf-8 bytes.  Therefore, there is no encoding
    parameter.
    """
    def __init__(self, autoflush=False):
        self._cdb = CDatabase()
        self.autoflush = autoflush

    def putDoc(self, doc, clean=False, weight=False):
        """
        Add a document to the index
        """
        flags = 0
        if clean:
            flags = flags | self._cdb.PDCLEAN
        if weight:
            flags = flags | self._cdb.PDWEIGHT
        if not self._cdb.put_doc(doc._cdoc, flags):
            msg = self._cdb.err_msg(self._cdb.error())
            raise PutFailed(doc[u'@uri'], msg)

        if self.autoflush:
            self.flush()

        return doc

    def flush(self):
        """
        Write documents to disk and break them into words
        """
        failed = False
        try:
            self._cdb.flush(-1)
        except CError:
            failed = True
        if failed:
            msg = self._cdb.err_msg(self._cdb.error())
            raise FlushFailed(msg)

    @unicodeToByte([(2,'uri')], TypeError)
    def remove(self, doc=None, uri=None, id=None):
        """
        Take a document out of the database by reference, by uri or by id
        """
        if uri is None and id is None and doc is None:
            raise TypeError("Either doc, uri or id is required to remove a document")

        flags = 0 # TODO - support ODCLEAN flag
        if hasattr(doc, 'id'):
            id = doc.id
        else:
            if uri is not None:
                id = self._cdb.uri_to_id(uri)
        if not self._cdb.out_doc(id, flags):
            msg = self._cdb.err_msg(self._cdb.error())
            raise EditFailed(id, msg)

    def __delitem__(self, uri):
        return self.remove(uri=uri)

    def updateAttributes(self, doc):
        """
        Edit a document's attributes in-place.  Note: there is no way to edit
        the texts.
        """
        failed = False
        try:
            self._cdb.edit_doc(doc._cdoc)
        except CError:
            failed = True
        if failed:
            msg = self._cdb.err_msg(self._cdb.error())
            raise EditFailed(doc.id, msg)

    def optimize(self, purge=False, opt=False):
        """
        Perform either a free-space compact operation, or a db optimization,
        or both, or neither, depending on flags.
        """
        flags = (0 if purge else self._cdb.OPTNOPURGE)
        flags = flags | (0 if opt else self._cdb.OPTNODBOPT)
        self._cdb.optimize(flags)

    def __len__(self):
        return self._cdb.doc_num()

    def sync(self):
        """
        ??? If anyone knows what this should do and how it is different from
        flush(), tell me.
        """
        raise NotImplementedError("I'll implement this when someone tells me what it does")
        self._cdb.sync() # TODO - needs a unit test

    def open(self, directory, mode):
        """
        Open the database directory.  Only valid modes are 'a', 'r' and 'w'

        'a' corresponds to WRITER | CREAT (created only if it does not exist)
        'w' corresponds to WRITER | CREAT | TRUNC (clobbered if it exists)
        'r' corresponds to READER
        """
        flags = 0
        if mode == 'a':
            flags = self._cdb.DBWRITER | self._cdb.DBCREAT
        elif mode == 'w':
            flags = self._cdb.DBWRITER | self._cdb.DBCREAT | self._cdb.DBTRUNC
        elif mode == 'r':
            flags = self._cdb.DBREADER

        assert flags, "mode must be 'a', 'w' or 'r'"

        if not self._cdb.open(directory, flags):
            msg = self._cdb.err_msg(self._cdb.error())
            raise OpenFailed(msg)

        return self

    def close(self):
        """
        Put the database down for the night.
        """
        failed = False
        try:
            self._cdb.close()
        except CError:
            failed = True
        if failed:
            msg = self._cdb.err_msg(self._cdb.error())
            raise CloseFailed(msg)

    def search(self, condition):
        """
        Submit a query to the database and return the results object.
        """
        result = self._cdb.search(condition.condition)
        return HResults(self, result)

    def walkResult(self, result):
        """
        Produce a HHit for each doc in the result
        """
        count = result.doc_num()
        for i in range(count):
            doc = self._cdb.get_doc(result.get_doc_id(i), 0) # TODO - flags

            if doc is None: # XXX WTF? is this for a race condition against
                continue    # someone removing docs from the index?

            yield HHit.fromCDocument(doc)

    @unicodeToByte([(1,'uri')], KeyError)
    def __getitem__(self, uri):
        id = self._cdb.uri_to_id(uri)
        return HDocument.fromCDocument(self._cdb.get_doc(id, 0))

    def __iter__(self):
        cond = HCondition()
        cond.addAttr(u'@id')
        return iter(self.search(cond))


class HResults(list):
    """
    List wrapper for results of the search.
    """
    def __init__(self, db, result):
        self._cresult = result
        list.__init__(self, db.walkResult(result))

    def hintWords(self):
        """
        Return the unicode-d search terms
        """
        return [w.decode('utf-8') for w in self._cresult.hint_words()]

    def pluck(self, attribute):
        """
        Return the value of the given attribute for each result, in a list
        """
        return [h[attribute] for h in self]


class HDocument(object):
    """
    Dict-like interface to a document

    >>> doc = HDocument(u'http://sample.com/page.html')
    """
    def __str__(self):
        return self._cdoc.dump_draft()

    @unicodeToByte([(1,'uri')], TypeError)
    def __init__(self, uri):
        self._cdoc = CDocument()
        self._cdoc.add_attr('@uri', uri.replace('\0', ''))

    @unicodeToByte([(1,'text')], TypeError)
    def addHiddenText(self, text):
        """
        Add text that will affect search scoring but will NOT appear in the
        output document
        """
        self._cdoc.add_hidden_text(text.replace('\0', ''))

    @unicodeToByte([(1,'text')], TypeError)
    def addText(self, text):
        """
        Put some text into the document
        """
        self._cdoc.add_text(text.replace('\0', ''))

    @classmethod
    def fromCDocument(cls, cdocument):
        """
        Construct a document from an existing estraiernative.Document, such as
        when iterating search results.
        """
        self = cls(uri=cdocument.attr('@uri').decode('utf8'))
        self._cdoc = cdocument
        return self

    def __delitem__(self, name):
        raise NotImplementedError("Cannot delete attributes from this document")

    @unicodeToByte([(1,'name'), (2,'value')], TypeError)
    def __setitem__(self, name, value):
        self._cdoc.add_attr(name.replace('\0', ''),
                value.replace('\0', ''))

    @unicodeToByte([(1,'name')], KeyError)
    def __getitem__(self, name):
        if name not in self._cdoc.attr_names():
            raise KeyError(name)
        return self._cdoc.attr(name).decode('utf-8')

    def update(self, other):
        """
        Update attributes of this document from another one
        """
        for k, v in other.items():
            self[k] = v

    def get(self, key, default=None):
        """
        Return doc[key] unless key is not found, in which case return the
        value of 'default' (None unless specified)
        """
        try:
            return self[key]
        except KeyError:
            return default

    def keys(self):
        """
        Names of all attributes set on this document
        """
        return [n.decode('utf-8') for n in self._cdoc.attr_names()]

    def values(self):
        """
        Values of all attributes set on this document
        """
        return [self[a] for a in self.keys()]

    def items(self):
        """
        (attribute_name, attribute_value) 2-tuples for every attribute in this
        document
        """
        return [(a,self[a]) for a in self.keys()]

    def getTexts(self):
        """
        Return all (visible) texts in this document, as a list
        """
        return map(lambda t: t.decode('utf-8'), self._cdoc.texts())

    def _get_text(self):
        return u'\n'.join(self.getTexts())
    text = property(_get_text)

    def _get_id(self):
        return self._cdoc.id()
    id = property(_get_id)

    def encode(self, encoding):
        """
        Return an encoded version of this document.  Convenience method
        """
        return ''.join([s.encode(encoding) for s in self.getTexts()])


class HHit(HDocument):
    """
    A hit returned by a search.
    """

    teaserFormats = 'html rst'.split()

    def emphasize_html(self, text):
        """
        Write emphasized text in html
        """
        return '<strong>%s</strong>' % text

    def emphasize_rst(self, text):
        """
        Write emphasized text in restructuredtext
        """
        return '**%s**' % text

    def teaser(self, terms, format='html'):
        """
        Produce the teaser/snippet text for the hit
        """
        if format not in self.teaserFormats:
            raise NotImplementedError( "Supported formats are: %s" % (
                ', '.join(self.teaserFormats),))

        emph = getattr(self, 'emphasize_%s' % (format,))

        # arguments to make_snippet MUST be byte strings
        bterms = []
        for t in terms:
            if not type(t) is type(u''):
                raise TypeError("Terms must be unicode text")
            bterms.append(t.encode('utf8'))

        # TODO - parameterize make_snippet's numbers
        snip = self._cdoc.make_snippet(bterms, 120, 35, 35)

        # parse the newline-delimited snippet format
        bunches = snip.split('\n\n')
        strings = []
        for bunch in snip.split('\n\n'):
            _bitStrings = []
            for bit in bunch.split('\n'):
                if '\t' in bit:
                    _bitStrings.append(emph(bit.split('\t')[0]))
                else:
                    _bitStrings.append(bit)
            strings.append(''.join(_bitStrings))

        decode = lambda s: unicode(s, 'utf8')
        snip = u' ... '.join(map(decode, strings))
        return snip



"""Hypy - Pythonic Hyper Estraier"""
from hypy.lib import (PutFailed,
    OpenFailed,
    CloseFailed,
    EditFailed,
    FlushFailed,
    HCondition,
    HDatabase,
    HResults,
    HDocument,
    HHit,
    )

from hypy.copyright import __version__



#
import glob
import os
import re

try:
    import setuptools
except ImportError:
    from distribute_setup import use_setuptools
    use_setuptools()

from setuptools import setup, Extension, find_packages

from pkg_resources import parse_version


VERRX = re.compile(r'((\d+\.)+\d)\.txt$')

def versionFromReleaseNotes():
    """
    Hypy's official version is the latest version of the release notes found
    in doc/release-notes.
    """
    thisDir = os.path.abspath(__file__).rsplit(os.sep, 1)[0]
    if thisDir == __file__:
        thisDir = ''

    versions = glob.glob(os.sep.join(
        [thisDir, 'doc', 'release-notes', '[0-9]*.txt']))
    if versions:
        return VERRX.search(max(versions, key=parse_version)).group(1)
    else:
        return 'VERSION_NOT_FOUND'


ext = Extension("_estraiernative",
                ["estraiernative.c"],
                libraries=["estraier"],
                include_dirs=["/usr/include/estraier", "/usr/include/qdbm"],
                )

setup(
        name="Hypy", 
        description='Pythonic wrapper for Hyper Estraier',
        author='Yusuke YOSHIDA',
        author_email='usk@nrgate.jp',
        maintainer='Cory Dodt',
        maintainer_email='pypi@spam.goonmill.org',
        url='http://goonmill.org/hypy/',
        download_url='http://hypy-source.goonmill.org/archive/tip.tar.gz',
        version=versionFromReleaseNotes(),
        ext_modules=[ext],
        zip_safe=False,
        packages=find_packages(),

        install_requires=[
            'Distribute>=0.6.3',
            ],

        classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries',
          'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
          ],
      )
